from unittest import mock

from django.contrib.contenttypes.models import ContentType

from api_fhir_r4.configurations import R4IdentifierConfig
from api_fhir_r4.tests import GenericTestMixin
from api_fhir_r4.tests.mixin import FhirConverterTestMixin
from insuree.models import Family
from insuree.test_helpers import create_test_insuree
from invoice.models import Bill, BillItem
from fhir.resources.R4B.invoice import Invoice as FHIRInvoice
from api_fhir_r4.utils.timeUtils import TimeUtils
from api_fhir_r4.tests.utils import get_connection_payload,get_or_create_user_api

class BillInvoiceTestMixin(GenericTestMixin, FhirConverterTestMixin):
    _TEST_BILL_STATUS = 'active'
    _TEST_BILL_UUID = '12345678-1234-1234-1234-123456789012'
    _TEST_BILL_CODE = 'TEST-CODE'
    _TEST_BILL_SUBJECT_TYPE = None
    _TEST_BILL_SUBJECT_TYPE_CODING = 'claim-batch'
    _TEST_BILL_THIRD_PARTY = None
    _TEST_BILL_THIRD_PARTY_UUID = '98765432-1234-1234-1234-123456789012'
    _TEST_BILL_DATE = TimeUtils.str_iso_to_date('2021-01-01')
    _TEST_BILL_TOTAL_NET = 10000.0
    _TEST_BILL_TOTAL_GROSS = 10000.0
    _TEST_BILL_CURRENCY = 'USD'
    _TEST_LINE_ITEM_CHARGE_ITEM = None
    _TEST_LINE_ITEM_CHARGE_ITEM_CODING = 'claim'
    _TEST_LiNE_ITEM_QUANTITY = 2
    _TEST_LINE_ITEM_UNIT_PRICE = 5000.0
    _TEST_LINE_ITEM_BASE_PRICE_COMPONENT_TYPE = 'base'
    _TEST_LINE_ITEM_DISCOUNT_COMPONENT_TYPE = 'discount'
    _TEST_LINE_ITEM_DISCOUNT = 0.1
    _TEST_LINE_ITEM_DEDUCTION_PRICE_COMPONENT_TYPE = 'deduction'
    _TEST_LINE_ITEM_DEDUCTION = 100
    _TEST_LINE_ITEM_DEDUCTION_FACTOR = 1
    _TEST_LINE_ITEM_TAX_PRICE_COMPONENT_TYPE = 'tax'
    _TEST_LINE_ITEM_TAX_RATE = 0.02
    user = None
    @classmethod
    def setUpTestData(cls):
        cls.user = get_or_create_user_api()
        cls._TEST_BILL_SUBJECT_TYPE = ContentType.objects.get(model__iexact='BatchRun')
        cls._TEST_LINE_ITEM_CHARGE_ITEM = ContentType.objects.get(model__iexact='Claim')

    def create_test_imis_instance(self):
        self._TEST_BILL_INSUREE = create_test_insuree()
        imis_bill = Bill()
        imis_bill.code = self._TEST_BILL_CODE
        
        imis_bill.subject_type = self._TEST_BILL_SUBJECT_TYPE
        imis_bill.thirdparty = self._TEST_BILL_INSUREE.family
        imis_bill.thirdparty.uuid = self._TEST_BILL_THIRD_PARTY_UUID
        imis_bill.date_bill = self._TEST_BILL_DATE
        imis_bill.amount_net = self._TEST_BILL_TOTAL_NET
        imis_bill.amount_total = self._TEST_BILL_TOTAL_GROSS
        imis_bill.currency_code = self._TEST_BILL_CURRENCY
        imis_bill.save(user=self.user)
        self._TEST_BILL_UUID  =imis_bill.id
        imis_bill_line_item = BillItem()
        imis_bill_line_item.bill = imis_bill
        imis_bill_line_item.line_type = self._TEST_LINE_ITEM_CHARGE_ITEM
        imis_bill_line_item.quantity = self._TEST_LiNE_ITEM_QUANTITY
        imis_bill_line_item.unit_price = self._TEST_LINE_ITEM_UNIT_PRICE
        imis_bill_line_item.discount = self._TEST_LINE_ITEM_DISCOUNT
        imis_bill_line_item.deduction = self._TEST_LINE_ITEM_DEDUCTION
        imis_bill_line_item.tax_rate = None
        imis_bill_line_item.code = "1"
        imis_bill_line_item.save(user=self.user)

        with mock.patch('django.db.models.fields.related_descriptors.create_reverse_many_to_one_manager') as mock_patch:
            class MockedManager(mock.MagicMock):
                def all(self):
                    return [imis_bill_line_item]

            mock_patch.return_value = MockedManager()
            # noinspection PyStatementEffect,PyUnresolvedReferences
            imis_bill.line_items_bill  # This call assigns mocked related manager to the model

        return imis_bill

    def verify_imis_instance(self, imis_obj):
        raise NotImplementedError('verify_imis_instance() not implemented')

    def create_test_fhir_instance(self):
        raise NotImplementedError('create_test_fhir_instance() not implemented')

    def verify_fhir_instance(self, fhir_obj):
        self.assertIs(type(fhir_obj), FHIRInvoice)
        self.assertEqual(fhir_obj.status, self._TEST_BILL_STATUS)
        self.verify_fhir_identifier(fhir_obj, R4IdentifierConfig.get_fhir_uuid_type_code(), self._TEST_BILL_UUID)
        self.verify_fhir_identifier(fhir_obj, R4IdentifierConfig.get_fhir_generic_type_code(), self._TEST_BILL_CODE)
        self.verify_fhir_coding_exists(fhir_obj.type.coding, self._TEST_BILL_SUBJECT_TYPE_CODING)
        self.assertTrue(self._TEST_BILL_THIRD_PARTY_UUID in fhir_obj.recipient.reference)
        self.assertEqual(fhir_obj.date, self._TEST_BILL_DATE)
        self.assertEqual(fhir_obj.totalNet.value, self._TEST_BILL_TOTAL_NET, 1e-10)
        self.assertEqual(fhir_obj.totalNet.currency, self._TEST_BILL_CURRENCY)
        self.assertEqual(fhir_obj.totalGross.value, self._TEST_BILL_TOTAL_GROSS, 1e-10)
        self.assertEqual(fhir_obj.totalGross.currency, self._TEST_BILL_CURRENCY)

        self.assertGreater(len(fhir_obj.lineItem), 0)
        self.verify_fhir_coding_exists(fhir_obj.lineItem[0].chargeItemCodeableConcept.coding,
                                       self._TEST_LINE_ITEM_CHARGE_ITEM_CODING)
        self.assertGreater(len(fhir_obj.lineItem[0].priceComponent), 0)
        for price_component in fhir_obj.lineItem[0].priceComponent:
            if price_component.type == self._TEST_LINE_ITEM_BASE_PRICE_COMPONENT_TYPE:
                self.verify_price_component(price_component, self._TEST_LiNE_ITEM_QUANTITY,
                                            self._TEST_LINE_ITEM_UNIT_PRICE * self._TEST_LiNE_ITEM_QUANTITY,
                                            self._TEST_LINE_ITEM_BASE_PRICE_COMPONENT_TYPE)
            elif price_component.type == self._TEST_LINE_ITEM_DISCOUNT_COMPONENT_TYPE:
                self.verify_price_component(price_component, self._TEST_LINE_ITEM_DISCOUNT,
                                            -self._TEST_LINE_ITEM_UNIT_PRICE * self._TEST_LiNE_ITEM_QUANTITY * (
                                                self._TEST_LINE_ITEM_DISCOUNT),
                                            self._TEST_LINE_ITEM_DISCOUNT_COMPONENT_TYPE)
            elif price_component.type == self._TEST_LINE_ITEM_DEDUCTION_PRICE_COMPONENT_TYPE:
                self.verify_price_component(price_component, self._TEST_LINE_ITEM_DEDUCTION_FACTOR,
                                            -self._TEST_LINE_ITEM_DEDUCTION,
                                            self._TEST_LINE_ITEM_DEDUCTION_PRICE_COMPONENT_TYPE)
            # FIXME tax rate is more complex, requires an object
            # elif price_component.type == self._TEST_LINE_ITEM_TAX_PRICE_COMPONENT_TYPE:
            #     self.verify_price_component(price_component, self._TEST_LINE_ITEM_TAX_RATE,
            #                                 ((self._TEST_LINE_ITEM_UNIT_PRICE * self._TEST_LiNE_ITEM_QUANTITY * (
            #                                         1 - self._TEST_LINE_ITEM_DISCOUNT))
            #                                  - self._TEST_LINE_ITEM_DEDUCTION) * self._TEST_LINE_ITEM_TAX_RATE,
            #                                 self._TEST_LINE_ITEM_TAX_PRICE_COMPONENT_TYPE)

    def verify_price_component(self, price_component, expected_factor, expected_amount, expected_type):
        self.assertGreater(len(price_component.extension), 0)
        self.assertEqual(price_component.extension[0].valueMoney.currency, self._TEST_BILL_CURRENCY)
        self.assertEqual(price_component.extension[0].valueMoney.value, self._TEST_LINE_ITEM_UNIT_PRICE)
        if expected_type:
            self.assertEqual(price_component.type, expected_type)
        self.assertEqual(float(price_component.factor), expected_factor, 1e-10)
        self.assertEqual(float(price_component.amount.value), expected_amount, 1e-10)



from django.utils.translation import gettext
from location.models import HealthFacility
from api_fhir_r4.configurations import R4IdentifierConfig, R4LocationConfig
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from fhir.resources.R4B.location import Location as FHIRLocation
from api_fhir_r4.models.imisModelEnums import ImisHfLevel
from api_fhir_r4.utils import DbManagerUtils


class LocationSiteConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_hf, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_location = FHIRLocation.construct()
        cls.build_fhir_physical_type(fhir_location, imis_hf)
        cls.build_fhir_pk(fhir_location, imis_hf, reference_type)
        cls.build_fhir_location_identifier(fhir_location, imis_hf)
        cls.build_fhir_location_name(fhir_location, imis_hf)
        cls.build_fhir_type(fhir_location, imis_hf)
        cls.build_fhir_part_of(fhir_location, imis_hf, reference_type)
        cls.mode = 'instance'
        return fhir_location

    @classmethod
    def to_imis_obj(cls, fhir_location, audit_user_id):
        errors = []
        fhir_location = FHIRLocation(**fhir_location)
        imis_hf = HealthFacility() 
        cls.build_imis_hf_identiftier(imis_hf, fhir_location, errors)
        cls.build_imis_hf_name(imis_hf, fhir_location, errors)
        cls.build_imis_hf_level(imis_hf, fhir_location, errors)
        cls.build_imis_parent_location_id(imis_hf, fhir_location, errors)            
        cls.check_errors(errors)
        return imis_hf

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_facility_id_type()

    @classmethod
    def get_reference_obj_uuid(cls, imis_hf: HealthFacility):
        return imis_hf.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_hf: HealthFacility):
        return imis_hf.id

    @classmethod
    def get_reference_obj_code(cls, imis_hf: HealthFacility):
        return imis_hf.code


    @classmethod
    def get_fhir_resource_type(cls):
        return FHIRLocation

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        return DbManagerUtils.get_object_or_none(HealthFacility, **cls.get_database_query_id_parameteres_from_reference(reference))

    @classmethod
    def build_fhir_location_identifier(cls, fhir_location, imis_hf):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_hf)
        fhir_location.identifier = identifiers

    @classmethod
    def build_fhir_location_code_identifier(cls, identifiers, imis_hf):
        if imis_hf is not None:
            identifier = cls.build_fhir_identifier(imis_hf.code,
                                                   R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                   R4IdentifierConfig.get_fhir_facility_id_type())
            identifiers.append(identifier)

    @classmethod
    def build_imis_hf_identiftier(cls, imis_hf, fhir_location, errors):
        value = cls.get_fhir_identifier_by_code(fhir_location.identifier,
                                                R4IdentifierConfig.get_fhir_facility_id_type())
        if value:
            imis_hf.code = value
        cls.valid_condition(imis_hf.code is None, gettext('Missing location code'), errors)

    @classmethod
    def build_fhir_location_name(cls, fhir_location, imis_hf):
        fhir_location.name = imis_hf.name

    @classmethod
    def build_imis_hf_name(cls, imis_hf, fhir_location, errors):
        name = fhir_location.name
        if not cls.valid_condition(name is None,
                                   gettext('Missing location `name` attribute'), errors):
            imis_hf.name = name

    @classmethod
    def build_fhir_physical_type(cls, fhir_location, imis_hf):
        code = R4LocationConfig.get_fhir_code_for_site()
        text = "site"
        fhir_location.physicalType = \
            cls.build_codeable_concept(code, R4LocationConfig.get_fhir_location_physical_type_system(), text=text)

    @classmethod
    def build_fhir_type(cls, fhir_location, imis_hf):
        code = ""
        text = ""
        if imis_hf.level == ImisHfLevel.HEALTH_CENTER.value:
            code = R4LocationConfig.get_fhir_code_for_health_center()
            text = "health-center"
        elif imis_hf.level == ImisHfLevel.DISPENSARY.value:
            code = R4LocationConfig.get_fhir_code_for_dispensary()
            text = "dispensary"
        elif imis_hf.level == ImisHfLevel.HOSPITAL.value:
            code = R4LocationConfig.get_fhir_code_for_hospital()
            text = "hospital"
        fhir_location.type = \
            [cls.build_codeable_concept(code, R4LocationConfig.get_fhir_location_site_type_system(), text=text)]

    @classmethod
    def build_imis_hf_level(cls, imis_hf, fhir_location, errors):
        code = fhir_location.type[0].coding[0].code
        if code == R4LocationConfig.get_fhir_code_for_hospital():
            imis_hf.level = ImisHfLevel.HOSPITAL.value
        elif code == R4LocationConfig.get_fhir_code_for_dispensary():
            imis_hf.level = ImisHfLevel.DISPENSARY.value
        elif code == R4LocationConfig.get_fhir_code_for_health_center():
            imis_hf.level = ImisHfLevel.HEALTH_CENTER.value
        cls.valid_condition(imis_hf.level is None, gettext('Missing location level'), errors)

    @classmethod
    def build_fhir_part_of(cls, fhir_location, imis_hf, reference_type):
        if imis_hf.location is not None:
            fhir_location.partOf = \
                LocationSiteConverter.build_fhir_resource_reference(
                    imis_hf.location,
                    'Location',
                    imis_hf.location.code,
                    reference_type)

    @classmethod
    def build_imis_parent_location_id(cls, imis_hf, fhir_location, errors):
        if fhir_location.partOf:
            parent_id = fhir_location.partOf
            if not cls.valid_condition(parent_id is None,
                                       gettext('Missing location `parent id` attribute'), errors):
                # get the imis location object, check if exists
                uuid_location = parent_id.identifier.value
                hf_location = Location.objects.filter(uuid__iexact=uuid_location)
                if hf_location:
                    hf_location = hf_location.first()
                    imis_hf.location = hf_location



from uuid import UUID

from insuree.models import Family, FamilyType
from location.models import Location
from location.test_helpers import create_test_village

from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import GroupConverter
from api_fhir_r4.mapping.groupMapping import GroupTypeMapping
from fhir.resources.R4B.extension import Extension
from fhir.resources.R4B.humanname import HumanName
from fhir.resources.R4B.identifier import Identifier
from fhir.resources.R4B.group import Group, GroupMember
from fhir.resources.R4B.reference import Reference
from api_fhir_r4.tests import GenericTestMixin

from insuree.test_helpers import create_test_insuree, create_test_family


class GroupTestMixin( GenericTestMixin):

    _TEST_LAST_NAME = "TEST_LAST_NAME"
    _TEST_OTHER_NAME = "TEST_OTHER_NAME"
    _TEST_INSUREE_CHFID = "TestChfId1"
    _TEST_POVERTY_STATUS = False
    _TEST_GROUP_TYPE = None
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_INSUREE_UUID = "7240daef-5f8f-4b0f-9042-b221e66f184a"
    _TEST_GROUP_UUID = "8e33033a-9f60-43ad-be3e-3bfeb992aae5"    
    _TEST_VILLAGE_CODE = "RTDTMTVT"
    _TEST_VILLAGE_UUID = "637f05cf-d8e8-4135-8250-7f01f01936bc"
    _TEST_VILLAGE_NAME = "TEST_NAME"
    _TEST_VILLAGE_TYPE = "V"
    test_insuree = None
    test_village = None
    test_group = None
    sub_str = {}
    @classmethod
    def setUpTestData(cls):
        cls._TEST_GROUP_TYPE = FamilyType.objects.get(code="H")
        cls.test_village =create_test_village(custom_props={
            'code': cls._TEST_VILLAGE_CODE,
            'uuid': cls._TEST_VILLAGE_UUID,
            'name': cls._TEST_VILLAGE_NAME
            })
        cls.test_insuree = create_test_insuree(with_family=False, 
                                               custom_props={'current_village':cls.test_village})
        cls.test_group = create_test_family(
            custom_props = {
            'location': cls.test_village,
            'address': cls._TEST_ADDRESS,
            'family_type': cls._TEST_GROUP_TYPE,
            'uuid': cls._TEST_GROUP_UUID
            }
        )
        cls.sub_str[cls._TEST_VILLAGE_UUID] = cls.test_village.uuid
        cls.sub_str[cls._TEST_GROUP_UUID] = cls.test_group.uuid
        cls.sub_str[cls._TEST_INSUREE_UUID] = cls.test_insuree.uuid
        cls.sub_str[cls._TEST_INSUREE_CHFID] = cls.test_insuree.chf_id
        
    
    
    def create_test_imis_instance(self):

        return self.test_group

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(str(self.test_insuree.chf_id), imis_obj.head_insuree.chf_id)
        self.assertEqual(self._TEST_ADDRESS, imis_obj.address)
        self.assertEqual(self._TEST_POVERTY_STATUS, imis_obj.poverty)
        self.assertEqual(self._TEST_GROUP_TYPE.code, imis_obj.family_type.code)

    def create_test_fhir_instance(self):
                

        fhir_family = {}
        fhir_family['actual'] = True
        fhir_family['type'] = "Person"
        fhir_family = Group(**fhir_family)

        name = HumanName.construct()
        name.family = self.test_insuree.last_name
        name.given = [self.test_insuree.other_names]
        name.use = "usual"

        identifiers = []
        chf_id = GroupConverter.build_fhir_identifier(
            self.test_insuree.chf_id,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code()
        )

        identifiers.append(chf_id)

        uuid = GroupConverter.build_fhir_identifier(
            self._TEST_GROUP_UUID,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_uuid_type_code()
        )
        identifiers.append(uuid)

        fhir_family.identifier = identifiers

        fhir_family.quantity = 1
        fhir_family.name = self.test_insuree.last_name

        members = []
        member = GroupMember.construct()
        reference = Reference.construct()
        reference.reference = f"Patient/{str(self.test_insuree.uuid)}"
        reference.type = "Patient"
        member.entity = reference
        members.append(member)
        fhir_family.member = members

        # extensions for group
        fhir_family.extension = []

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-poverty-status"
        extension.valueBoolean = self._TEST_POVERTY_STATUS
        fhir_family.extension.append(extension)

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-type"
        GroupTypeMapping.load()
        display = GroupTypeMapping.group_type[str(self._TEST_GROUP_TYPE.code)]
        system = f"CodeSystem/group-type"
        extension.valueCodeableConcept = GroupConverter.build_codeable_concept(code=str(self._TEST_GROUP_TYPE.code), system=system)
        if len(extension.valueCodeableConcept.coding) == 1:
            extension.valueCodeableConcept.coding[0].display = display
        fhir_family.extension.append(extension)

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-address"
        family_address = GroupConverter.build_fhir_address(self._TEST_ADDRESS, "home", "physical")
        family_address.state = self.test_village.parent.parent.parent.name
        family_address.district = self.test_village.parent.parent.name

        # municipality extension
        extension_address = Extension.construct()
        extension_address.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-municipality"
        extension_address.valueString = self.test_village.parent.name
        family_address.extension = [extension_address]

        # address location reference extension
        extension_address = Extension.construct()
        extension_address.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-location-reference"
        reference_location = Reference.construct()
        reference_location.reference = F"Location/{self.test_village.uuid}"
        extension_address.valueReference = reference_location
        family_address.extension.append(extension_address)
        family_address.city = self.test_village.name
        extension.valueAddress = family_address

        fhir_family.extension.append(extension)
        self.test_group = fhir_family
        return self.test_group

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual('Person', fhir_obj.type)
        self.assertEqual(True, fhir_obj.actual)
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = GroupConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_generic_type_code():
                pass#FIXME fail chf-id form DB self.assertEqual(str(self.test_insuree.chf_id), identifier.value)
            elif code == R4IdentifierConfig.get_fhir_uuid_type_code() and not isinstance(identifier.value, UUID):
                self.assertEqual(str(self.test_group.uuid), identifier.value)
        if False: #FIXME
            self.assertEqual(1, fhir_obj.quantity)
            self.assertEqual(1, len(fhir_obj.member))
            self.assertEqual(self._TEST_LAST_NAME, fhir_obj.name)
            self.assertEqual(3, len(fhir_obj.extension))
            for extension in fhir_obj.extension:
                self.assertTrue(isinstance(extension, Extension))
                if "group-address" in extension.url:
                    no_of_extensions = len(extension.valueAddress.extension)
                    self.assertEqual(2, no_of_extensions)
                    self.assertEqual("home", extension.valueAddress.use)
                    self.assertEqual(self._TEST_VILLAGE_NAME, extension.valueAddress.city)
                if "group-poverty-status" in extension.url:
                    self.assertEqual(self._TEST_POVERTY_STATUS, extension.valueBoolean)
                if "group-type" in extension.url:
                    self.assertEqual(self._TEST_GROUP_TYPE.code, extension.valueCodeableConcept.coding[0].code)



from django.test import TestCase

import core
import dateutil

from api_fhir_r4.utils import TimeUtils


class TimeUtilsTestCase(TestCase):
    __OFFSET_IN_SECONDS = 3600

    def test_str_converting_datetime(self):
        str_value = "2010-11-16T15:22:01"
        expected = core.datetime.datetime(2010, 11, 16, 15, 22, 1, 0)
        actual = TimeUtils.str_iso_to_date(str_value)
        self.assertEqual(expected, actual)

    def test_str_converting_datetime_with_time_zone(self):
        str_value = "2010-11-16T15:22:01+01:00"
        expected = core.datetime.datetime(2010, 11, 16, 15, 22, 1, 0,
                                          tzinfo=dateutil.tz.tzoffset(None, self.__OFFSET_IN_SECONDS))
        actual = TimeUtils.str_iso_to_date(str_value)
        self.assertEqual(expected, actual)

    def test_str_converting_date(self):
        str_value = "2010-11-16"
        expected = core.datetime.date(2010, 11, 16)
        actual = TimeUtils.str_to_date(str_value)
        self.assertEqual(expected, actual)

    def test_str_converting_date_and_time(self):
        str_date_value = "2010-11-16"
        str_time_value = "01:02:03"
        expected = core.datetime.datetime(2010, 11, 16, 1, 2, 3)
        actual = TimeUtils.str_to_date(str_date_value, str_time_value)
        self.assertEqual(expected, actual)



from api_fhir_r4.converters import ClaimAdminPractitionerConverter

from fhir.resources.R4B.practitioner import Practitioner
from api_fhir_r4.tests import ClaimAdminPractitionerTestMixin
from api_fhir_r4.tests.mixin import ConvertToImisTestMixin, ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin


class ClaimAdminPractitionerConverterTestCase(ClaimAdminPractitionerTestMixin,
                                              ConvertToImisTestMixin,
                                              ConvertToFhirTestMixin,
                                              ConvertJsonToFhirTestMixin):
    converter = ClaimAdminPractitionerConverter
    fhir_resource = Practitioner
    json_repr = 'test/test_claimAdminPractitioner.json'



import core
import uuid

from policy.models import Policy
from insuree.models import InsureePolicy

from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig, R4CoverageConfig
from api_fhir_r4.converters import ContractConverter
from api_fhir_r4.tests import GenericTestMixin

from django.utils.translation import gettext as _
from fhir.resources.R4B.contract import Contract, ContractTerm, ContractTermOffer, \
    ContractTermAsset, ContractTermOfferParty, ContractTermAssetValuedItem
from fhir.resources.R4B.identifier import Identifier
from fhir.resources.R4B.money import Money
from fhir.resources.R4B.period import Period
from fhir.resources.R4B.reference import Reference
from api_fhir_r4.utils import TimeUtils

from core.test_helpers import create_test_officer
from insuree.test_helpers import create_test_insuree
from product.test_helpers import create_test_product


class ContractTestMixin(GenericTestMixin):

    _TEST_POLICY_ID = "100000"
    _TEST_POLICY_ENROLL_DATE = "2021-06-20T00:00:00"
    _TEST_POLICY_START_DATE = "2021-06-20T00:00:00"
    _TEST_POLICY_EFFECTIVE_DATE = "2021-06-20T00:00:00"
    _TEST_POLICY_EXPIRED_DATE = "2022-06-19T00:00:00"
    _TEST_POLICY_STATUS = 1
    _TEST_POLICY_STAGE = 'N'
    _TEST_GROUP_UUID = "e8bbb7e4-19ef-4bef-9342-9ab6b9a928d3"
    _TEST_OFFICER_UUID = "ff7db42d-874b-400a-bba7-e59b273ae123"
    _TEST_INSUREE_UUID = "f8c56ada-d76d-4f6c-aad3-cfddc9fb38eb"
    _TEST_PRODUCT_CODE = "TE123"
    _TEST_PRODUCT_UUID = "8ed8d2d9-2644-4d29-ba37-ab772386cfca"

    _TEST_POLICY_VALUE = 10000.0

    def create_test_imis_instance(self):
        imis_policy = Policy()
        imis_policy.id = self._TEST_POLICY_ID

        # create mocked insuree
        imis_insuree = create_test_insuree(with_family=True)
        imis_insuree.uuid = self._TEST_INSUREE_UUID
        imis_insuree.save()

        # update family uuid
        imis_family = imis_insuree.family
        imis_family.uuid = self._TEST_GROUP_UUID
        imis_family.save()

        # create mocked product
        imis_product = create_test_product(self._TEST_PRODUCT_CODE, valid=True, custom_props=None)
        imis_product.uuid = self._TEST_PRODUCT_UUID
        imis_product.save()

        imis_policy.family = imis_family
        imis_policy.product = imis_product

        imis_policy.enroll_date = TimeUtils.str_to_date(self._TEST_POLICY_ENROLL_DATE)
        imis_policy.start_date = TimeUtils.str_to_date(self._TEST_POLICY_START_DATE)
        imis_policy.effective_date = TimeUtils.str_to_date(self._TEST_POLICY_EFFECTIVE_DATE)
        imis_policy.expiry_date = TimeUtils.str_to_date(self._TEST_POLICY_EXPIRED_DATE)

        imis_policy.stage = self._TEST_POLICY_STAGE
        imis_policy.status = self._TEST_POLICY_STATUS
        imis_policy.value = self._TEST_POLICY_VALUE
        imis_policy.audit_user_id = -1

        # create mocked officer
        imis_officer = create_test_officer()
        imis_officer.uuid = self._TEST_OFFICER_UUID
        imis_officer.save()
        imis_policy.officer = imis_officer

        # save mock policy
        imis_policy.save()

        # create mock policy insuree
        imis_policy_insuree = InsureePolicy(
            policy=imis_policy,
            insuree=imis_insuree,
            audit_user_id=-1
        )
        imis_policy_insuree.save()

        return imis_policy

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_POLICY_ENROLL_DATE, imis_obj.enroll_date.isoformat())
        self.assertEqual(self._TEST_POLICY_START_DATE, imis_obj.start_date.isoformat())
        self.assertEqual(self._TEST_POLICY_EFFECTIVE_DATE, imis_obj.effective_date.isoformat())
        self.assertEqual(self._TEST_POLICY_EXPIRED_DATE, imis_obj.expiry_date.isoformat())
        self.assertEqual(self._TEST_GROUP_UUID, str(uuid.UUID(imis_obj.family.uuid)))
        self.assertEqual(self._TEST_OFFICER_UUID, str(uuid.UUID(imis_obj.officer.uuid)))
        self.assertEqual(self._TEST_PRODUCT_CODE, imis_obj.product.code)
        self.assertEqual(self._TEST_PRODUCT_UUID, str(uuid.UUID(imis_obj.product.uuid)))

    def create_test_fhir_instance(self):
        # create some dependency
        # create mocked insuree
        imis_insuree = create_test_insuree(with_family=True)
        imis_insuree.uuid = self._TEST_INSUREE_UUID
        imis_insuree.save()

        # create mocked family
        imis_family = imis_insuree.family
        imis_family.uuid = self._TEST_GROUP_UUID
        imis_family.save()

        # create mocked product
        imis_product = create_test_product(self._TEST_PRODUCT_CODE, valid=True, custom_props=None)
        imis_product.uuid = self._TEST_PRODUCT_UUID
        imis_product.save()

        # create mocked officer
        imis_officer = create_test_officer()
        imis_officer.uuid = self._TEST_OFFICER_UUID
        imis_officer.save()

        if hasattr(core, 'currency'):
            currency = core.currency
        else:
            currency = "EUR"

        fhir_contract = Contract.construct()
        id = ContractConverter.build_fhir_identifier(
            self._TEST_POLICY_ID,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_id_type_code()
        )
        identifiers = [id]
        fhir_contract.identifier = identifiers

        author = Reference.construct()
        author.reference = f"Practitioner/{self._TEST_OFFICER_UUID}"
        fhir_contract.author = author

        subject = Reference.construct()
        subject.reference = f"Group/{self._TEST_GROUP_UUID}"
        fhir_contract.subject = [subject]

        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/contract-scope"
        fhir_contract.scope = ContractConverter.build_codeable_concept(code="informal", system=system)
        if len(fhir_contract.scope.coding) == 1:
            fhir_contract.scope.coding[0].display = _("Informal Sector")

        # contract term build
        contract_term = ContractTerm.construct()

        # contract term offer
        offer = ContractTermOffer.construct()
        offer_party = ContractTermOfferParty.construct()
        reference = Reference.construct()
        insuree_uuid = self._TEST_INSUREE_UUID
        reference.reference = f"Patient/{insuree_uuid}"
        offer_party.reference = [reference]
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/contract-resource-party-role"
        offer_party.role = ContractConverter.build_codeable_concept(code="beneficiary", system=system)
        if len(offer_party.role.coding) == 1:
            offer_party.role.coding[0].display = _("Beneficiary")

        offer.party = [offer_party]
        contract_term.offer = offer

        contract_term_asset = ContractTermAsset.construct()
        # contract term asset extensions - premium

        typeReference = Reference.construct()
        typeReference.reference = f"Patient/{insuree_uuid}"
        contract_term_asset.typeReference = [typeReference]

        valued_item = ContractTermAssetValuedItem.construct()
        typeReference = Reference.construct()
        typeReference.reference = f"InsurancePlan/{self._TEST_PRODUCT_UUID}"
        valued_item.entityReference = typeReference
        policy_value = Money.construct()
        policy_value.value = self._TEST_POLICY_VALUE
        valued_item.net = policy_value
        contract_term_asset.valuedItem = [valued_item]

        period_use = Period.construct()
        period = Period.construct()
        #if imis_policy.effective_date is not None:
        period_use.start = self._TEST_POLICY_EFFECTIVE_DATE
        #if period_use.start is None:
        period.start = period_use.start
        #if imis_policy.expiry_date is not None:
        period_use.end = self._TEST_POLICY_EXPIRED_DATE
        period.end = period_use.end

        contract_term_asset.usePeriod = [period_use]
        contract_term_asset.period = [period]

        contract_term.asset = [contract_term_asset]
        fhir_contract.term = [contract_term]

        fhir_contract.status = R4CoverageConfig.get_status_offered_code()
        fhir_contract.legalState = ContractConverter.build_simple_codeable_concept(
            R4CoverageConfig.get_status_offered_code()
        )
        return fhir_contract

    def verify_fhir_instance(self, fhir_obj):
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = ContractConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_id_type_code():
                self.assertEqual(self._TEST_POLICY_ID, identifier.value)
        self.assertIn(f"Group/{self._TEST_GROUP_UUID}", fhir_obj.subject[0].reference)
        self.assertIn(f"Practitioner/{self._TEST_OFFICER_UUID}", fhir_obj.author.reference)
        self.assertEqual("Offered", fhir_obj.status)
        self.assertEqual("Offered", fhir_obj.legalState.text)
        term = fhir_obj.term[0]
        offer = term.offer
        asset = term.asset[0]
        reference_asset = asset.typeReference[0].reference
        reference_asset = reference_asset.split('Patient/')[1].lower()
        self.assertEqual(self._TEST_INSUREE_UUID, reference_asset)
        period = asset.period[0]
        self.assertEqual(self._TEST_POLICY_START_DATE, period.start.isoformat()+"T00:00:00")
        self.assertEqual(self._TEST_POLICY_EXPIRED_DATE, period.end.isoformat()+"T00:00:00")
        use_period = asset.usePeriod[0]
        self.assertEqual(self._TEST_POLICY_EFFECTIVE_DATE, use_period.start.isoformat()+"T00:00:00")
        self.assertEqual(self._TEST_POLICY_EXPIRED_DATE, use_period.end.isoformat()+"T00:00:00")
        valued_item = asset.valuedItem[0]
        self.assertIn(f"InsurancePlan/{self._TEST_PRODUCT_UUID}", valued_item.entityReference.reference)
        self.assertEqual(self._TEST_POLICY_VALUE, valued_item.net.value)



# Generated by Django 3.2.16 on 2022-10-12 08:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api_fhir_r4', '0004_update_subscription_criteria'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='historicalsubscription',
            options={'get_latest_by': ('history_date', 'history_id'), 'ordering': ('-history_date', '-history_id'), 'verbose_name': 'historical subscription', 'verbose_name_plural': 'historical subscriptions'},
        ),
        migrations.AlterField(
            model_name='historicalsubscription',
            name='history_date',
            field=models.DateTimeField(db_index=True),
        ),
    ]



import copy

from .utils import PaymentNoticeSerializerUtils
from invoice.services import PaymentInvoiceService


class PaymentNoticeSerializerCreate(object):
    @classmethod
    def create(cls, validated_data, request):
        user = request.user

        imis_payment_detail = validated_data.pop('imis_payment_detail')
        invoice_status = validated_data.pop('invoice_status')

        payment_invoice = copy.deepcopy(validated_data)
        del payment_invoice['_state']
        del payment_invoice['_original_state']

        payment_invoice_service = PaymentInvoiceService(user)
        result = payment_invoice_service.create_with_detail(payment_invoice, imis_payment_detail)
        payment_invoice = PaymentNoticeSerializerUtils.get_result_object(result)
        if invoice_status and result['success']:
            # change the invoice status based on 'fhir status'
            PaymentNoticeSerializerUtils.change_invoice_status(payment_invoice_service, payment_invoice)
        return payment_invoice



import json
from pathlib import Path

from django.test import TestCase
from api_fhir_r4.tests.utils import load_and_replace_json


class FhirConverterTestMixin(TestCase):
    def verify_fhir_identifier(self, fhir_obj, identifier_type, expected_identifier_value):
        identifiers = [identifier for identifier in fhir_obj.identifier
                       if identifier.type.coding[0].code == identifier_type]
        self.assertEqual(len(identifiers), 1)
        self.assertEqual(str(identifiers[0].value), str(expected_identifier_value))

    def verify_fhir_coding_exists(self, fhir_coding, expected_code):
        self.assertIsNotNone(next(iter([coding for coding in fhir_coding if coding.code == expected_code]), None))


class ConvertToImisTestMixin:
    @property
    def converter(self):
        raise NotImplementedError()

    def create_test_fhir_instance(self):
        raise NotImplementedError()

    def verify_imis_instance(self, imis_instance):
        raise NotImplementedError()

    def test_to_imis_obj(self):
        fhir_instance = self.create_test_fhir_instance()
        imis_instance = self.converter.to_imis_obj(fhir_instance.dict(), None)
        self.verify_imis_instance(imis_instance)


class ConvertToFhirTestMixin:
    @property
    def converter(self):
        raise NotImplementedError()

    def create_test_imis_instance(self):
        raise NotImplementedError()

    def verify_fhir_instance(self, fhir_instance):
        raise NotImplementedError()

    def test_to_fhir_obj(self):
        imis_instance = self.create_test_imis_instance()
        fhir_instance = self.converter.to_fhir_obj(imis_instance)
        self.verify_fhir_instance(fhir_instance)


class ConvertJsonToFhirTestMixin:
    sub_str = {}
    @property
    def fhir_resource(self):
        raise NotImplementedError()

    @property
    def json_repr(self):
        raise NotImplementedError()

    def verify_fhir_instance(self, fhir_instance):
        raise NotImplementedError()

    def test_fhir_object_from_json(self):
        fhir_instance = self.fhir_resource.parse_obj(load_and_replace_json(self.json_repr, self.sub_str))
        self.verify_fhir_instance(fhir_instance)





from api_fhir_r4.configurations import OrganisationConfiguration


class R4OrganisationConfig(OrganisationConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_organisation_config = cfg['R4_fhir_organisation_config']

    @classmethod
    def get_fhir_ph_organisation_type(cls):
        return cls.get_config_attribute('R4_fhir_organisation_config').get('fhir_ph_organisation_type', 'bus')

    @classmethod
    def get_fhir_ph_organisation_type_system(cls):
        return cls.get_config_attribute('R4_fhir_organisation_config') \
            .get('fhir_ph_organisation_type_system',
                 'http://terminology.hl7.org/CodeSystem/organization-type')

    @classmethod
    def get_fhir_ph_organisation_contactentity_type(cls):
        return cls.get_config_attribute('R4_fhir_organisation_config') \
            .get('fhir_ph_organisation_type_contactentity_type',
                 'http://terminology.hl7.org/CodeSystem/contactentity-type')

    @classmethod
    def get_fhir_address_municipality_extension_system(cls):
        return cls.get_config_attribute('R4_fhir_organisation_config').get('fhir_address_municipality_extension_system',
                                                                           'StructureDefinition/address-municipality')

    @classmethod
    def get_fhir_location_reference_extension_system(cls):
        return cls.get_config_attribute('R4_fhir_organisation_config') \
            .get('fhir_location_reference_extension_system',
                 'StructureDefinition/address-location-reference')

    @classmethod
    def get_fhir_ph_organisation_legal_form_extension_system(cls):
        return cls.get_config_attribute('R4_fhir_organisation_config') \
            .get('fhir_ph_organisation_legal_form_extension_system',
                 'StructureDefinition/organization-ph-legal-form')

    @classmethod
    def get_fhir_ph_organisation_activity_extension_system(cls):
        return cls.get_config_attribute('R4_fhir_organisation_config') \
            .get('fhir_ph_organisation_activity_extension_system',
                 'StructureDefinition/organization-ph-activity')

    @classmethod
    def get_fhir_ph_organisation_legal_form_code_system(cls):
        return cls.get_config_attribute('R4_fhir_organisation_config') \
            .get('fhir_ph_organisation_legal_form_code_system',
                 'CodeSystem/organization-ph-legal-form')

    @classmethod
    def get_fhir_ph_organisation_activity_code_system(cls):
        return cls.get_config_attribute('R4_fhir_organisation_config') \
            .get('fhir_ph_organisation_activity_code_system',
                 'CodeSystem/organization-ph-activity')

    @classmethod
    def get_fhir_insurer_organisation_id(cls):
        return cls.get_config_attribute('R4_fhir_insurance_organisation_config') \
            .get('id', 'openIMIS-Implementation')

    @classmethod
    def get_fhir_insurer_organisation_code(cls):
        return cls.get_config_attribute('R4_fhir_insurance_organisation_config') \
            .get('code', 'openIMIS')

    @classmethod
    def get_fhir_insurer_organisation_name(cls):
        return cls.get_config_attribute('R4_fhir_insurance_organisation_config') \
            .get('name', 'openIMIS Implementation')

    @classmethod
    def get_fhir_insurer_organisation_type(cls):
        return cls.get_config_attribute('R4_fhir_insurance_organisation_config') \
            .get('type', 'ins')

    @classmethod
    def get_fhir_insurer_organisation_email(cls):
        return cls.get_config_attribute('R4_fhir_insurance_organisation_config') \
            .get('email', 'implementation@openimis.org')

    @classmethod
    def get_fhir_insurer_organisation_phone(cls):
        return cls.get_config_attribute('R4_fhir_insurance_organisation_config') \
            .get('phone', '0908060703')

    @classmethod
    def get_fhir_insurer_organisation_fax(cls):
        return cls.get_config_attribute('R4_fhir_insurance_organisation_config') \
            .get('fax', '0908060730')

    @classmethod
    def get_fhir_insurer_organisation_contact_name(cls):
        return cls.get_config_attribute('R4_fhir_insurance_organisation_config') \
            .get('contact_name', 'Manuel D. Medina')

    @classmethod
    def get_fhir_insurer_organisation_municipality(cls):
        return cls.get_config_attribute('R4_fhir_insurance_organisation_config') \
            .get('municipality', 'Jamu')

    @classmethod
    def get_fhir_insurer_organisation_city(cls):
        return cls.get_config_attribute('R4_fhir_insurance_organisation_config') \
            .get('city', 'Jamula')

    @classmethod
    def get_fhir_insurer_organisation_district(cls):
        return cls.get_config_attribute('R4_fhir_insurance_organisation_config') \
            .get('district', 'Rapta')

    @classmethod
    def get_fhir_insurer_organisation_state(cls):
        return cls.get_config_attribute('R4_fhir_insurance_organisation_config') \
            .get('state', 'Ultha')

    @classmethod
    def get_fhir_insurer_organisation_line(cls):
        return cls.get_config_attribute('R4_fhir_insurance_organisation_config') \
            .get('line', '1 Pasay')



from claim.models import ClaimAdmin
from django.utils.translation import gettext

from api_fhir_r4.converters import ClaimAdminPractitionerRoleConverter
from api_fhir_r4.exceptions import FHIRRequestProcessException
from api_fhir_r4.serializers import BaseFHIRSerializer


class ClaimAdminPractitionerRoleSerializer(BaseFHIRSerializer):

    fhirConverter = ClaimAdminPractitionerRoleConverter()

    def create(self, validated_data):
        claim_admin_id = validated_data.get('id')
        claim_admin = None
        if claim_admin_id:
            claim_admin = ClaimAdmin.objects.get(pk=claim_admin_id)
        if not claim_admin:
            raise FHIRRequestProcessException([gettext('Missing Practitioner for id {}').format(claim_admin_id)])
        claim_admin.health_facility_id = validated_data.get('health_facility_id', claim_admin.health_facility_id)
        claim_admin.save()
        return claim_admin

    def update(self, instance, validated_data):
        instance.health_facility_id = validated_data.get('health_facility_id', instance.health_facility_id)
        instance.audit_user_id = self.get_audit_user_id()
        instance.save()
        return instance



import asyncio
import decimal
import json
import logging
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass

import aiohttp

from typing import Union, Dict, List, Any, TypeVar, Generic, Iterable

import orjson

from api_fhir_r4.models import Subscription

NOTIFICATION_CONTENT_TYPE = TypeVar('NOTIFICATION_CONTENT_TYPE')  # FHIR INPUT
CLIENT_ACCEPTABLE_CONTENT_TYPE = TypeVar('CLIENT_ACCEPTABLE_CONTENT_TYPE')  # CLIENT INPUT
NOTIFICATION_OUTPUT_TYPE = TypeVar('NOTIFICATION_OUTPUT_TYPE')  # CLIENT RESPONSE

logger = logging.getLogger('openIMIS')


class AbstractAsyncSubscriptionNotificationClient(
        Generic[NOTIFICATION_CONTENT_TYPE, CLIENT_ACCEPTABLE_CONTENT_TYPE, NOTIFICATION_OUTPUT_TYPE], ABC):
    def propagate_notifications(self, notification_content: NOTIFICATION_CONTENT_TYPE, subscribers: List[Subscription])\
            -> Iterable[NOTIFICATION_OUTPUT_TYPE]:
        """
        Create new asyncio event loop and call propagate_notifications_async.

        Args:
            notification_content: Message to be sent to all recipients
            subscribers: List of recipients, eligible for receiving given notification.

        Returns:
            List of responses or errors occurred during notifying subscribers
        """
        return asyncio.run(self.propagate_notifications_async(notification_content, subscribers))

    async def propagate_notifications_async(self, content: NOTIFICATION_CONTENT_TYPE, subscribers: List[Subscription])\
            -> Iterable[NOTIFICATION_OUTPUT_TYPE]:
        payload = self._normalize_payload(content)
        async with aiohttp.ClientSession() as session:
            tasks = []
            for sub in subscribers:
                task = asyncio.ensure_future(self._send_notification_async(payload, sub, session))
                tasks.append(task)
            result = await asyncio.gather(*tasks)
            return result

    @abstractmethod
    def _normalize_payload(self, payload: NOTIFICATION_CONTENT_TYPE) -> CLIENT_ACCEPTABLE_CONTENT_TYPE:
        """
        Transforms payload to format accepted by _send_notification_async

        Args:
            payload: Content of notification being sent to subscribers

        Returns:
            Payload transformed to type acceptable by client.
        """
        pass

    @abstractmethod
    async def _send_notification_async(
            self, content: CLIENT_ACCEPTABLE_CONTENT_TYPE, subscriber: Subscription,
            client_session: aiohttp.ClientSession) -> NOTIFICATION_OUTPUT_TYPE:
        """
        Uses client_session for sending content to designated subscriber.

        @param content: Content of notification
        @param subscriber: Recipient of the notification
        @param client_session: Open client session responsible for sending content.
        @return:
        """
        pass


RestNotificationContentType = Union[Dict, str]


@dataclass
class SubscriberNotificationOutput:
    subscription: Subscription
    notification_success: bool
    reason_of_failure: Any = None


class RestSubscriptionNotificationClient(AbstractAsyncSubscriptionNotificationClient[
                                RestNotificationContentType, Union[str, bytes], SubscriberNotificationOutput]):
    async def _send_notification_async(self, content: CLIENT_ACCEPTABLE_CONTENT_TYPE, subscriber: Subscription,
                                       client_session: aiohttp.ClientSession) -> NOTIFICATION_OUTPUT_TYPE:
        try:
            post_args = self._post_args(content, subscriber)
            async with client_session.post(**post_args) as post:
                response = await post.json()
                status = post.status
                if status >= 400:
                    return SubscriberNotificationOutput(subscriber, False, response)
                else:
                    return SubscriberNotificationOutput(subscriber, True, None)
        except Exception as e:
            logger.error(F"Sending subscription notification has failed due to {e}")
            import traceback
            logger.debug(traceback.format_exc())
            return SubscriberNotificationOutput(subscriber, False, e)

    def _normalize_payload(self, payload: NOTIFICATION_CONTENT_TYPE) -> CLIENT_ACCEPTABLE_CONTENT_TYPE:
        return payload if isinstance(payload, str) else self.__transform_payload(payload)

    @property
    def _base_headers(self):
        return {
            'content-type': 'application/json',
            'accept': 'application/json'
        }

    @staticmethod
    def __transform_payload(payload):
        def uuid_convert(o):
            if isinstance(o, uuid.UUID):
                return o.hex
            if isinstance(o, decimal.Decimal):
                return float(o)
        return orjson.dumps(payload, default=uuid_convert)

    def _post_args(self, content, subscriber: Subscription):
        try:
            subscriber_headers = json.loads(subscriber.headers)
        except TypeError as e:
            logger.debug(f"Notification failed due to invalid headers format: {subscriber.headers}.")
            raise TypeError(f"Invalid format of headers for '{subscriber}'."
                            f" Headers should be provided as JSON string.") \
                from e
        return {
            'headers': {**self._base_headers, **subscriber_headers},
            'url': subscriber.endpoint,
            'data': content
        }



from claim.models import Claim, ClaimItem, ClaimService
from insuree.test_helpers import create_test_insuree
from medical.models import Diagnosis

from api_fhir_r4.configurations import R4IdentifierConfig, R4ClaimConfig
from api_fhir_r4.converters import PatientConverter, HealthFacilityOrganisationConverter, \
    ClaimAdminPractitionerConverter, ReferenceConverterMixin
from api_fhir_r4.converters.claimConverter import ClaimConverter
from api_fhir_r4.mapping.claimMapping import ClaimVisitTypeMapping
from api_fhir_r4.models import ClaimV2 as FHIRClaim
from fhir.resources.R4B.period import Period
from fhir.resources.R4B.money import Money
from location.models import HealthFacility
from location.test_helpers import create_test_village, create_test_health_facility
from medical.models import Item, Service
from medical.test_helpers import create_test_item, create_test_service
from claim.test_helpers import create_test_claimservice,create_test_claimitem,create_test_claim_admin
from api_fhir_r4.tests import GenericTestMixin
from api_fhir_r4.utils import TimeUtils


class ClaimTestMixin(GenericTestMixin):
    _TEST_UUID = "315c3b16-62eb-11ea-8e75-df3492b349f6"
    _TEST_CLAIM_CODE = 'T00001'
    _TEST_DATE_FROM = TimeUtils.str_to_date('2021-02-03')
    _TEST_DATE_TO = TimeUtils.str_to_date('2021-02-03')

    # diagnosis data
    _TEST_MAIN_ICD_CODE = 'T_CD'
    _TEST_MAIN_ICD_NAME = 'Test diagnosis'

    _TEST_CLAIMED = 21100
    _TEST_DATE_CLAIMED = '2021-02-03T00:00:00'
    _TEST_GUARANTEE_ID = "guarantee_id"
    _TEST_EXPLANATION = "example explanation"
    _TEST_VISIT_TYPE = "O"
    _TEST_STATUS = Claim.STATUS_REJECTED

    # claim item data
    _TEST_ITEM_CODE = "0004"
    _TEST_ITEM_UUID = "e2bc1546-390b-4d41-8571-632ecf7a0936"
    _TEST_ITEM_QUANTITY_PROVIDED = 10.0
    _TEST_ITEM_PRICE_ASKED = 10.0
    _TEST_ITEM_EXPLANATION = "item_explanation"
    _TEST_ITEM_TYPE = 'D'

    # claim service data
    _TEST_SERVICE_CODE = "M7"
    _TEST_SERVICE_UUID = "a17602f4-e9ff-4f42-a6a4-ccefcb23b4d6"
    _TEST_SERVICE_QUANTITY_PROVIDED = 1
    _TEST_SERVICE_PRICE_ASKED = 21000.0
    _TEST_SERVICE_EXPLANATION = "service_explanation"
    _TEST_SERVICE_TYPE = 'D'

    # hf test data
    _TEST_HF_ID = 10000
    _TEST_HF_UUID = "6d0eea8c-62eb-11ea-94d6-c36229a16c2f"
    _TEST_HF_CODE = "12345678"
    _TEST_HF_NAME = "TEST_NAME"
    _TEST_HF_LEVEL = "H"
    _TEST_HF_LEGAL_FORM = "G"
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_PHONE = "133-996-476"
    _TEST_FAX = "1-408-999 8888"
    _TEST_EMAIL = "TEST@TEST.com"

    # insuree and claim admin data
    _TEST_INSUREE_UUID = "76aca309-f8cf-4890-8f2e-b416d78de00b"
    _TEST_CLAIM_ADMIN_UUID = "044c33d1-dbf3-4d6a-9924-3797b461e535"

    _ADMIN_AUDIT_USER_ID = -1
    test_insuree = None
    test_icd = None
    test_hf = None
    test_claim_admin = None
    test_claim = None
    test_claim_item = None
    test_claim_service = None
    sub_str = {}

    def setUp(self):
        super(ClaimTestMixin, self).setUp()
        self.test_icd = Diagnosis()
        self.test_icd.code = self._TEST_MAIN_ICD_CODE
        self.test_icd.name = self._TEST_MAIN_ICD_NAME
        self.test_icd.audit_user_id = self._ADMIN_AUDIT_USER_ID
        self.test_icd.save()

        self.test_claim_admin= create_test_claim_admin()
        self.test_insuree = create_test_insuree()

        self.create_test_hf()

        self.test_claim=self.create_test_claim()
        self.test_claim_item= self.create_test_claim_item()
        self.test_claim_service = self.create_test_claim_service()
        self.sub_str[self._TEST_HF_UUID]=self.test_hf.uuid
        self.sub_str[self._TEST_CLAIM_ADMIN_UUID]=self.test_claim_admin.uuid
        self.sub_str[self._TEST_INSUREE_UUID]=self.test_insuree.uuid
        self.sub_str[self._TEST_SERVICE_UUID]=self.test_claim_service.service.uuid
        self.sub_str[self._TEST_ITEM_UUID]=self.test_claim_item.item.uuid
        self.sub_str[self._TEST_SERVICE_CODE]=self.test_claim_service.service.code
        self.sub_str[self._TEST_ITEM_CODE]=self.test_claim_item.item.code
        self.sub_str[self._TEST_UUID]=self.test_claim.uuid
        self._TEST_HF_ID = self.test_hf.id
        self._TEST_HF_UUID = self.test_hf.uuid
        
    def create_test_hf(self):
        self.test_hf = create_test_health_facility(
            self._TEST_HF_CODE,
            self.test_insuree.family.location.parent.parent.id,
            custom_props = {
                'name': self._TEST_HF_NAME,
                'level':self._TEST_HF_LEVEL,
                'legal_form_id':self._TEST_HF_LEGAL_FORM,
                'address':self._TEST_ADDRESS,
                'phone':self._TEST_PHONE,
                'fax':self._TEST_FAX,
                'email':self._TEST_EMAIL,
            }
        )
        return self.test_hf
    def create_test_claim(self):
        imis_claim = Claim()
        imis_claim.uuid = self._TEST_UUID
        imis_claim.insuree = self.test_insuree
        imis_claim.code = self._TEST_CLAIM_CODE
        imis_claim.date_from = TimeUtils.str_to_date(self._TEST_DATE_FROM)
        imis_claim.date_to = TimeUtils.str_to_date(self._TEST_DATE_TO)

        imis_claim.icd = self.test_icd

        imis_claim.claimed = self._TEST_CLAIMED
        imis_claim.date_claimed = TimeUtils.str_to_date(self._TEST_DATE_CLAIMED)
        imis_claim.health_facility = self.test_hf
        imis_claim.guarantee_id = self._TEST_GUARANTEE_ID
        imis_claim.admin = self.test_claim_admin
        imis_claim.visit_type = self._TEST_VISIT_TYPE
        imis_claim.status = self._TEST_STATUS
        imis_claim.audit_user_id = self._ADMIN_AUDIT_USER_ID
        imis_claim.save()
        return imis_claim
    
    def create_test_claim_item(self):
        item = Item.objects.filter(code=self._TEST_ITEM_CODE).first()
        if item is None:
            item = create_test_item(
                self._TEST_ITEM_TYPE,
                custom_props={"code": self._TEST_ITEM_CODE}
            )
        return create_test_claimitem( self.test_claim, self._TEST_ITEM_TYPE,
            custom_props={
                "item": item,
                "price_asked": self._TEST_ITEM_PRICE_ASKED,
                "qty_provided": self._TEST_ITEM_QUANTITY_PROVIDED,
                "explanation": self._TEST_ITEM_EXPLANATION,
                "audit_user_id": self._ADMIN_AUDIT_USER_ID
            }
        )
    
        
    def create_test_claim_service(self):
        service = Service.objects.filter(code=self._TEST_ITEM_CODE).first()
        if service is None:
            service = create_test_service( 
                self._TEST_SERVICE_TYPE,
                custom_props={"code": self._TEST_ITEM_CODE}
            )
        return create_test_claimservice(self.test_claim, self._TEST_SERVICE_TYPE,
            custom_props={
                "service": service,
                "price_asked": self._TEST_SERVICE_PRICE_ASKED,
                "qty_provided": self._TEST_SERVICE_QUANTITY_PROVIDED,
                "explanation": self._TEST_SERVICE_EXPLANATION,
                "audit_user_id": self._ADMIN_AUDIT_USER_ID
            }
        )

    def create_test_imis_instance(self):
        return self.test_claim

    def verify_imis_instance(self, imis_obj):
        self.assertIsNotNone(imis_obj.insuree)
        self.assertEqual(self._TEST_CLAIM_CODE, imis_obj.code)
        self.assertEqual(self._TEST_DATE_FROM.isoformat(), imis_obj.date_from.isoformat())
        self.assertEqual(self._TEST_DATE_TO.isoformat(), imis_obj.date_to.isoformat())
        self.assertEqual(self._TEST_MAIN_ICD_CODE, imis_obj.icd.code)
        self.assertEqual(self._TEST_CLAIMED, imis_obj.claimed)
        self.assertEqual(self._TEST_DATE_CLAIMED, imis_obj.date_claimed.isoformat())
        self.assertIsNotNone(imis_obj.health_facility)
        self.assertEqual(self._TEST_GUARANTEE_ID, imis_obj.guarantee_id)
        self.assertEqual(self._TEST_EXPLANATION, imis_obj.explanation)
        self.assertIsNotNone(imis_obj.admin)
        self.assertEqual(self._TEST_VISIT_TYPE, imis_obj.visit_type)

        self.assertEqual(self.test_claim_item.item.code, imis_obj.submit_items[0].item.code)
        #FIXME self.assertEqual(self._TEST_ITEM_QUANTITY_PROVIDED, imis_obj.submit_items[0].qty_provided)
        #self.assertEqual(self._TEST_ITEM_PRICE_ASKED, imis_obj.submit_items[0].price_asked)

        self.assertEqual(self.test_claim_service.service.code, imis_obj.submit_services[0].service.code)
        #self.assertEqual(self._TEST_SERVICE_QUANTITY_PROVIDED, imis_obj.submit_services[0].qty_provided)
        #self.assertEqual(self._TEST_SERVICE_PRICE_ASKED, imis_obj.submit_services[0].price_asked)

    def create_test_fhir_instance(self):
        fhir_claim = {}
        fhir_claim["use"] = 'claim'
        fhir_claim["status"] = 'active'
        fhir_claim["created"] = self._TEST_DATE_CLAIMED
        fhir_claim = FHIRClaim(**fhir_claim)

        mapping = ClaimVisitTypeMapping.fhir_claim_visit_type_coding[self._TEST_VISIT_TYPE]
        fhir_claim.type = ClaimConverter.build_codeable_concept_from_coding(
            ClaimConverter.build_fhir_mapped_coding(mapping))

        fhir_claim.patient = PatientConverter.build_fhir_resource_reference(self.test_insuree)
        claim_code = ClaimConverter.build_fhir_identifier(
            self._TEST_CLAIM_CODE,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_claim_code_type()
        )
        fhir_claim.identifier = [claim_code]

        billable_period = Period.construct()
        billable_period.start = self._TEST_DATE_FROM
        billable_period.end = self._TEST_DATE_TO
        fhir_claim.billablePeriod = billable_period

        diagnoses = []
        ClaimConverter.build_fhir_diagnosis(
            diagnoses,
            self.test_icd,
        )
        fhir_claim.diagnosis = diagnoses

        supportingInfo = []
        guarantee_id_code = R4ClaimConfig.get_fhir_claim_information_guarantee_id_code()
        ClaimConverter.build_fhir_string_information(supportingInfo, guarantee_id_code, self._TEST_GUARANTEE_ID)
        explanation_code = R4ClaimConfig.get_fhir_claim_information_explanation_code()
        ClaimConverter.build_fhir_string_information(supportingInfo, explanation_code, self._TEST_EXPLANATION)

        fhir_claim.supportingInfo = supportingInfo

        fhir_claim.enterer = ClaimAdminPractitionerConverter.build_fhir_resource_reference(
            self.test_claim_admin
        )

        fhir_claim.item = []
        type = R4ClaimConfig.get_fhir_claim_item_code()
        ClaimConverter.build_fhir_item(fhir_claim, self._TEST_ITEM_CODE, type, self.test_claim_item,
                                       reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE)
        type = R4ClaimConfig.get_fhir_claim_service_code()
        ClaimConverter.build_fhir_item(fhir_claim, self._TEST_SERVICE_CODE, type, self.test_claim_service,
                                       reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE)

        fhir_claim.provider = HealthFacilityOrganisationConverter.build_fhir_resource_reference(
            self.test_hf,
            display=self._TEST_HF_CODE
        )

        total = Money.construct()
        total.value = self._TEST_CLAIMED
        fhir_claim.total = total

        return fhir_claim

    def verify_fhir_instance(self, fhir_obj):
        self.assertIsNotNone(fhir_obj.patient.reference)
        for identifier in fhir_obj.identifier:
            if identifier.type.coding[0].code == R4IdentifierConfig.get_fhir_claim_code_type():
                self.assertEqual(self._TEST_CLAIM_CODE, identifier.value)

        self.assertIn(fhir_obj.billablePeriod.start.isoformat(), self._TEST_DATE_FROM.isoformat())
        self.assertIn(fhir_obj.billablePeriod.end.isoformat(), self._TEST_DATE_TO.isoformat())
        for diagnosis in fhir_obj.diagnosis:
            code = diagnosis.diagnosisCodeableConcept.coding[0].code
            self.assertEqual(self._TEST_MAIN_ICD_CODE, code)

        self.assertEqual(self._TEST_CLAIMED, fhir_obj.total.value)
        self.assertIn(fhir_obj.created.isoformat(), self._TEST_DATE_CLAIMED)
        for supportingInfo in fhir_obj.supportingInfo:
            if supportingInfo.category.text == R4ClaimConfig.get_fhir_claim_information_explanation_code():
                self.assertEqual(self._TEST_EXPLANATION, supportingInfo.valueString)
            elif supportingInfo.category.text == R4ClaimConfig.get_fhir_claim_information_guarantee_id_code():
                self.assertEqual(self._TEST_GUARANTEE_ID, supportingInfo.valueString)
        self.assertIsNotNone(fhir_obj.provider.reference)
        self.assertIn(str(self.test_hf.uuid), fhir_obj.provider.reference)
        self.assertIsNotNone(fhir_obj.enterer.reference)
        self.assertIn(str(self.test_claim_admin.uuid), fhir_obj.enterer.reference)
        self.assertIsNotNone(fhir_obj.patient.reference)
        self.assertIn(str(self.test_insuree.uuid), fhir_obj.patient.reference)
        for item in fhir_obj.item:
            if item.category.text == R4ClaimConfig.get_fhir_claim_item_code():
                self.assertEqual(self.test_claim_item.item.code, item.productOrService.text)
                self.assertEqual(self._TEST_ITEM_QUANTITY_PROVIDED, item.quantity.value)
                self.assertEqual(self._TEST_ITEM_PRICE_ASKED, item.unitPrice.value)
            elif item.category.text == R4ClaimConfig.get_fhir_claim_service_code():
                return None#FIXME
                self.assertEqual(self.test_claim_service.service.code, item.productOrService.text)
                self.assertEqual(self._TEST_SERVICE_QUANTITY_PROVIDED, item.quantity.value)
                self.assertEqual(self._TEST_SERVICE_PRICE_ASKED, item.unitPrice.value)



import json
import os

from api_fhir_r4.converters import InsurancePlanConverter

from fhir.resources.R4B.insuranceplan import InsurancePlan
from api_fhir_r4.tests import InsurancePlanTestMixin
from api_fhir_r4.tests.mixin import ConvertToImisTestMixin, ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin


class InsurancePlanConverterTestCase(InsurancePlanTestMixin,
                                     ConvertToImisTestMixin,
                                     ConvertToFhirTestMixin,
                                     ConvertJsonToFhirTestMixin):
    converter = InsurancePlanConverter
    fhir_resource = InsurancePlan
    json_repr = 'test/test_insurance_plan.json'



from copy import deepcopy
from urllib import parse
import datetime
from fhir.resources.R4B.subscription import Subscription as FHIRSubscription

from api_fhir_r4.configurations import R4SubscriptionConfig
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.mapping.subscriptionMapping import SubscriptionChannelMapping, SubscriptionStatusMapping
from api_fhir_r4.models import Subscription
from api_fhir_r4.utils import TimeUtils


class SubscriptionConverter(BaseFHIRConverter):
    _error_unknown_imis_value = f'Unknown imis `%(attr)s`: %(val)s'
    _error_invalid_attr = f'Missing or invalid `%(attr)s` attribute'
    _error_forbidden_attr = f'`%(attr)s` attribute forbidden'

    @classmethod
    def to_fhir_obj(cls, imis_subscription, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_subscription = {}
        cls._build_fhir_id(fhir_subscription, imis_subscription)
        cls._build_fhir_status(fhir_subscription, imis_subscription)
        cls._build_fhir_end(fhir_subscription, imis_subscription)
        cls._build_fhir_reason(fhir_subscription, imis_subscription)
        cls._build_fhir_criteria(fhir_subscription, imis_subscription)
        cls._build_fhir_error(fhir_subscription, imis_subscription)
        cls._build_fhir_channel(fhir_subscription, imis_subscription)
        return FHIRSubscription.parse_obj(fhir_subscription)

    @classmethod
    def to_imis_obj(cls, fhir_subscription, audit_user_id):
        fhir_subscription = FHIRSubscription.parse_obj(fhir_subscription)
        imis_subscription = {}
        cls._build_imis_id(imis_subscription, fhir_subscription)
        cls._build_imis_status(imis_subscription, fhir_subscription)
        cls._build_imis_end(imis_subscription, fhir_subscription)
        cls._build_imis_criteria(imis_subscription, fhir_subscription)
        cls._validate_fhir_error(fhir_subscription)
        cls._build_imis_channel_type(imis_subscription, fhir_subscription)
        cls._build_imis_channel_endpoint(imis_subscription, fhir_subscription)
        cls._build_imis_channel_header(imis_subscription, fhir_subscription)
        return Subscription(**imis_subscription)

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return FHIRException("Subscription resource does not contain code identifier")

    @classmethod
    def _build_fhir_id(cls, fhir_subscription, imis_subscription):
        fhir_subscription['id'] = str(imis_subscription.id)

    @classmethod
    def _build_fhir_status(cls, fhir_subscription, imis_subscription):
        if imis_subscription.status in SubscriptionStatusMapping.to_fhir_status:
            fhir_subscription['status'] = SubscriptionStatusMapping.to_fhir_status[imis_subscription.status]
        else:
            raise FHIRException(
                cls._error_unknown_imis_value % {'attr': 'status', 'val': str(imis_subscription.status)})

    @classmethod
    def _build_fhir_end(cls, fhir_subscription, imis_subscription):
        date = imis_subscription.expiring.astimezone().isoformat()
        fhir_subscription['end'] = date

    @classmethod
    def _build_fhir_reason(cls, fhir_subscription, imis_subscription):
        fhir_subscription['reason'] = \
            imis_subscription.criteria[R4SubscriptionConfig.get_fhir_sub_criteria_key_resource()]

    @classmethod
    def _build_fhir_criteria(cls, fhir_subscription, imis_subscription):
        criteria = deepcopy(imis_subscription.criteria)
        resource = criteria.pop(R4SubscriptionConfig.get_fhir_sub_criteria_key_resource())
        resource_type = criteria.pop(R4SubscriptionConfig.get_fhir_sub_criteria_key_resource_type(), None)
        if resource_type:
            criteria['resourceType'] = resource_type
        fhir_subscription['criteria'] = \
            parse.urlunparse(('', '', resource, '', parse.urlencode(criteria, doseq=True), ''))

    @classmethod
    def _build_fhir_error(cls, fhir_subscription, imis_subscription):
        error = None  # TODO Add that after the error logging is added to the model
        if error:
            fhir_subscription['error'] = error

    @classmethod
    def _build_fhir_channel(cls, fhir_subscription, imis_subscription):
        fhir_channel = {}
        cls._build_fhir_channel_type(fhir_channel, imis_subscription)
        cls._build_fhir_channel_endpoint(fhir_channel, imis_subscription)
        cls._build_fhir_channel_header(fhir_channel, imis_subscription)
        fhir_subscription['channel'] = fhir_channel

    @classmethod
    def _build_fhir_channel_type(cls, fhir_channel, imis_subscription):
        if imis_subscription.channel in SubscriptionChannelMapping.to_fhir_channel:
            fhir_channel['type'] = SubscriptionChannelMapping.to_fhir_channel[imis_subscription.channel]
        else:
            raise FHIRException(
                cls._error_unknown_imis_value % {'attr': 'channel', 'val': str(imis_subscription.channel)})

    @classmethod
    def _build_fhir_channel_endpoint(cls, fhir_channel, imis_subscription):
        fhir_channel['endpoint'] = imis_subscription.endpoint

    @classmethod
    def _build_fhir_channel_header(cls, fhir_channel, imis_subscription):
        fhir_channel['header'] = [imis_subscription.headers]

    @classmethod
    def _build_imis_id(cls, imis_subscription, fhir_subscription):
        if fhir_subscription.id:
            imis_subscription['id'] = fhir_subscription.id

    @classmethod
    def _build_imis_status(cls, imis_subscription, fhir_subscription):
        if fhir_subscription.status and fhir_subscription.status in SubscriptionStatusMapping.to_imis_status:
            imis_subscription['status'] = SubscriptionStatusMapping.to_imis_status[fhir_subscription.status]
        else:
            raise FHIRException(cls._error_invalid_attr % {'attr': 'status'})

    @classmethod
    def _build_imis_end(cls, imis_subscription, fhir_subscription):
        if fhir_subscription.end:
            imis_subscription['expiring'] = TimeUtils.str_iso_to_date(fhir_subscription.end).astimezone(
                datetime.timezone.utc)
        else:
            raise FHIRException(cls._error_invalid_attr % {'attr': 'end'})

    @classmethod
    def _build_imis_criteria(cls, imis_subscription, fhir_subscription):
        if fhir_subscription.criteria:
            parsed_criteria = parse.urlparse(fhir_subscription.criteria)
            imis_criteria = dict(parse.parse_qsl(parsed_criteria.query))
            resource_type = imis_criteria.pop('resourceType', None)
            if resource_type:
                imis_criteria[R4SubscriptionConfig.get_fhir_sub_criteria_key_resource_type()] = resource_type
            imis_criteria[R4SubscriptionConfig.get_fhir_sub_criteria_key_resource()] = parsed_criteria.path.strip('/')
            imis_subscription['criteria'] = imis_criteria
            cls._validate_fhir_reason(imis_subscription, fhir_subscription)
        else:
            raise FHIRException(cls._error_invalid_attr % {'attr': 'criteria'})

    @classmethod
    def _validate_fhir_reason(cls, imis_subscription, fhir_subscription):
        # there is no imis reason so validation only
        if 'criteria' not in imis_subscription:
            raise FHIRException(cls._error_invalid_attr % {'attr': 'criteria'})
        if fhir_subscription.reason.lower() \
                != imis_subscription['criteria'][R4SubscriptionConfig.get_fhir_sub_criteria_key_resource()].lower():
            raise FHIRException(cls._error_invalid_attr % {'attr': 'reason'})

    @classmethod
    def _validate_fhir_error(cls, fhir_subscription):
        # there should not be error attribute in fhir input
        if fhir_subscription.error:
            raise FHIRException(cls._error_forbidden_attr % {'attr': 'error'})

    @classmethod
    def _build_imis_channel_type(cls, imis_subscription, fhir_subscription):
        if fhir_subscription.channel.type \
                and fhir_subscription.channel.type in SubscriptionChannelMapping.to_imis_channel:
            imis_subscription['channel'] = SubscriptionChannelMapping.to_imis_channel[fhir_subscription.channel.type]
        else:
            raise FHIRException(cls._error_invalid_attr % {'attr': 'channel.type'})

    @classmethod
    def _build_imis_channel_endpoint(cls, imis_subscription, fhir_subscription):
        if fhir_subscription.channel.endpoint:
            imis_subscription['endpoint'] = str(fhir_subscription.channel.endpoint)
        else:
            raise FHIRException(cls._error_invalid_attr % {'attr': 'channel.endpoint'})

    @classmethod
    def _build_imis_channel_header(cls, imis_subscription, fhir_subscription):
        if fhir_subscription.channel.header:
            imis_subscription['headers'] = fhir_subscription.channel.header[0]



from typing import Union

from django.db.models import Q

from api_fhir_r4.configurations import R4SubscriptionConfig
from api_fhir_r4.models import Subscription
from core.datetimes.ad_datetime import datetime
from core.models import HistoryModel, VersionedModel


class SubscriptionCriteriaFilter:
    def __init__(self, imis_resource: Union[HistoryModel, VersionedModel], fhir_resource_name: str,
                 fhir_resource_type_name: str):
        self.fhir_resource_name = fhir_resource_name
        self.fhir_resource_type_name = fhir_resource_type_name
        self.imis_resource = imis_resource

    def get_filtered_subscriptions(self):
        subscriptions = self._get_all_active_subscriptions()
        return self._get_matching_subscriptions(subscriptions)

    def _get_all_active_subscriptions(self):
        queryset = Subscription.objects.filter(status=Subscription.SubscriptionStatus.ACTIVE.value,
                                               expiring__gt=datetime.now(), is_deleted=False)
        if self.fhir_resource_name:
            queryset = queryset.filter(criteria__jsoncontains={
                R4SubscriptionConfig.get_fhir_sub_criteria_key_resource(): self.fhir_resource_name})
        if self.fhir_resource_type_name:
            queryset = queryset.filter(
                ~Q(criteria__jsoncontainskey=R4SubscriptionConfig.get_fhir_sub_criteria_key_resource_type() )| Q(
                    criteria__jsoncontains={
                        R4SubscriptionConfig.get_fhir_sub_criteria_key_resource_type(): self.fhir_resource_type_name}))
        return queryset.all()

    def _get_matching_subscriptions(self, subscriptions):
        return [subscription for subscription in subscriptions
                if self._is_matching_subscription(subscription)]

    def _is_matching_subscription(self, sub):
        criteria = {criteria: sub.criteria[criteria] for criteria in sub.criteria if
                    criteria != R4SubscriptionConfig.get_fhir_sub_criteria_key_resource()
                    and criteria != R4SubscriptionConfig.get_fhir_sub_criteria_key_resource_type()}
        return not criteria or self._is_resource_matching_criteria(criteria)

    def _is_resource_matching_criteria(self, criteria):
        criteria['uuid'] = self.imis_resource.uuid
        return type(self.imis_resource).objects.filter(**criteria).exists()



import os
import json
from core.models import User
from api_fhir_r4.utils import DbManagerUtils
from core.services import create_or_update_interactive_user, create_or_update_core_user

def load_and_replace_json(path=None,sub_str={}):
    test_root = os.path.dirname(os.path.realpath(__file__))
    #removing the leading / of path
    path = path[1:] if path[0] == '/' else path
    json_representation = open(os.path.join(test_root,path)).read()

    for s_str in sub_str:
        json_representation = json_representation.replace(str(s_str),str(sub_str[s_str]))   
    
    return json.loads(json_representation)

# example of use
#        self.sub_str[self._TEST_VILLAGE_UUID] = self.test_village.uuid
#        self.sub_str[self._TEST_SUBSTITUTION_OFFICER_UUID] = self.test_officer.uuid
#        self.sub_str[self._TEST_GROUP_UUID] = self.test_insuree.family.uuid
#        self.sub_str[self._TEST_INSUREE_UUID] = self.test_insuree.uuid
#        self.sub_str[self._TEST_INSUREE_CHFID] = self.test_insuree.chf_id
#        self.sub_str[self._TEST_OFFICER_UUID] = self.test_officer.uuid
#        self.sub_str[self._TEST_CLAIM_ADMIN_UUID] = self.test_claim_admin.uuid
#        self.sub_str[self._TEST_PRODUCT_UUID] = self.test_product.uuid
#       
#        self.sub_str[self._TEST_HF_UUID] = self.test_hf.uuid
#        
#        self.json_representation = load_and_replace_json(_test_json_request_path,self.sub_str)
_TEST_USER_NAME = "TestUserTest2"
_TEST_USER_PASSWORD = "TestPasswordTest2"
_TEST_DATA_USER = {
    "username": _TEST_USER_NAME,
    "last_name": _TEST_USER_NAME,
    "password": _TEST_USER_PASSWORD,
    "other_names": _TEST_USER_NAME,
    "user_types": "INTERACTIVE",
    "language": "en",
    "roles": [1, 3, 5, 9],
}

def get_connection_payload(userdata= _TEST_DATA_USER):
    return {
  "username": userdata['username'],
  "password": userdata['password']
}


def get_or_create_user_api( userdata =_TEST_DATA_USER ):
    user = DbManagerUtils.get_object_or_none(User, username=userdata['username'])
    if user is None:
        user = __create_user_interactive_core(userdata)
    return user

def __create_user_interactive_core(userdata):
    i_user, i_user_created = create_or_update_interactive_user(
        user_id=None, data=userdata, audit_user_id=999, connected=False
    )
    create_or_update_core_user(
        user_uuid=None, username=userdata['username'], i_user=i_user
    )
    return DbManagerUtils.get_object_or_none(User, username=userdata['username'])


from api_fhir_r4.configurations import InvoiceConfiguration


class R4InvoiceConfig(InvoiceConfiguration):
    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_invoice_config = cfg['R4_fhir_invoice_config']

    @classmethod
    def get_fhir_invoice_type_system(cls):
        return cls.get_config_attribute("R4_fhir_invoice_config").get('fhir_invoice_type_system',
                                                                      "CodeSystem/invoice-type")

    @classmethod
    def get_fhir_invoice_charge_item_system(cls):
        return cls.get_config_attribute("R4_fhir_invoice_config").get('fhir_invoice_charge_item_system',
                                                                      "CodeSystem/invoice-charge-item")

    @classmethod
    def get_fhir_bill_type_system(cls):
        return cls.get_config_attribute("R4_fhir_invoice_config").get('fhir_bill_type_system',
                                                                      "CodeSystem/bill-type")

    @classmethod
    def get_fhir_bill_charge_item_system(cls):
        return cls.get_config_attribute("R4_fhir_invoice_config").get('fhir_bill_charge_item_system',
                                                                      "CodeSystem/bill-charge-item")

    @classmethod
    def get_subscribe_invoice_signal(cls):
        return cls.get_config_attribute("R4_fhir_invoice_config").get('subscribe_invoice_signal', False)



from api_fhir_r4.configurations import IssueTypeConfiguration


class R4IssueTypeConfig(IssueTypeConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_issue_type_config = cfg['R4_fhir_issue_type_config']

    @classmethod
    def get_fhir_code_for_exception(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_exception', 'exception')

    @classmethod
    def get_fhir_code_for_not_found(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_not_found', 'not-found')

    @classmethod
    def get_fhir_code_for_informational(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_informational', 'informational')



from api_fhir_r4.converters import CommunicationConverter
from api_fhir_r4.tests import CommunicationTestMixin
from fhir.resources.R4B.communication import Communication

from api_fhir_r4.tests.mixin import ConvertToImisTestMixin, ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin


class CommunicationConverterTestCase(CommunicationTestMixin,
                                     ConvertToImisTestMixin,
                                     ConvertToFhirTestMixin,
                                     ConvertJsonToFhirTestMixin):
    converter = CommunicationConverter
    fhir_resource = Communication
    json_repr = 'test/test_communication.json'



from api_fhir_r4.tests.mixin.genericTestMixin import GenericTestMixin
from api_fhir_r4.tests.mixin.locationTestMixin import LocationTestMixin
from api_fhir_r4.tests.mixin.operationOutcomeTestMixin import OperationOutcomeTestMixin
from api_fhir_r4.tests.mixin.patientTestMixin import PatientTestMixin
from api_fhir_r4.tests.mixin.claimAdminPractitionerTestMixin import ClaimAdminPractitionerTestMixin
from api_fhir_r4.tests.mixin.claimAdminPractitionerRoleTestMixin import ClaimAdminPractitionerRoleTestMixin
from api_fhir_r4.tests.mixin.claimTestMixin import ClaimTestMixin
from api_fhir_r4.tests.mixin.claimResponseTestMixin import ClaimResponseTestMixin
from api_fhir_r4.tests.mixin.activityDefinitionTestMixin import ActivityDefinitionTestMixin
from api_fhir_r4.tests.mixin.coverageEligibilityRequestTestMixin import CoverageEligibilityRequestTestMixin
from api_fhir_r4.tests.mixin.communicationRequestTestMixin import CommunicationRequestTestMixin
from api_fhir_r4.tests.mixin.communicationTestMixin import CommunicationTestMixin
from api_fhir_r4.tests.mixin.genericFhirAPITestMixin import GenericFhirAPITestMixin
from api_fhir_r4.tests.mixin.fhirApiReadTestMixin import FhirApiReadTestMixin
from api_fhir_r4.tests.mixin.fhirApiCreateTestMixin import FhirApiCreateTestMixin
from api_fhir_r4.tests.mixin.fhirApiUpdateTestMixin import FhirApiUpdateTestMixin
from api_fhir_r4.tests.mixin.fhirApiDeleteTestMixin import FhirApiDeleteTestMixin
from api_fhir_r4.tests.mixin.insurancePlanTestMixin import InsurancePlanTestMixin
from api_fhir_r4.tests.mixin.medicationTestMixin import MedicationTestMixin
from api_fhir_r4.tests.mixin.contractTestMixin import ContractTestMixin
from api_fhir_r4.tests.mixin.groupTestMixin import GroupTestMixin
from api_fhir_r4.tests.mixin.enrolmentOfficerPractitionerTestMixin import EnrolmentOfficerPractitionerTestMixin
from api_fhir_r4.tests.mixin.enrolmentOfficerPractitionerRoleTestMixin import EnrolmentOfficerPractitionerRoleTestMixin
from api_fhir_r4.tests.mixin.coverageTestMixin import CoverageTestMixin
from api_fhir_r4.tests.mixin.paymentNoticeTestMixin import PaymentNoticeTestMixin
from api_fhir_r4.tests.testCodeSystem import *



import logging
from collections import OrderedDict

import yaml
from django.apps import AppConfig

from api_fhir_r4.configurations import ModuleConfiguration
from api_fhir_r4.defaultConfig import DEFAULT_CFG

logger = logging.getLogger(__name__)

MODULE_NAME = "api_fhir_r4"


class ApiFhirConfig(AppConfig):
    name = MODULE_NAME

    def ready(self):
        from core.models import ModuleConfiguration
        cfg = ModuleConfiguration.get_or_default(MODULE_NAME, DEFAULT_CFG)
        self.__configure_module(cfg)
        setup_yaml()

        from openIMIS.ExceptionHandlerRegistry import ExceptionHandlerRegistry
        from .exceptions.fhir_api_exception_handler import fhir_api_exception_handler
        ExceptionHandlerRegistry.register_exception_handler(MODULE_NAME, fhir_api_exception_handler)

    def __configure_module(self, cfg):
        ModuleConfiguration.build_configuration(cfg)
        logger.info(F'Module {MODULE_NAME} configured successfully')


def setup_yaml():
    def represent_ordered_dict(dumper, data):
        return dumper.represent_mapping('tag:yaml.org,2002:map', data.items())

    yaml.SafeDumper.add_representer(OrderedDict, represent_ordered_dict)



from api_fhir_r4.configurations import R4PaymentNoticeConfig
from invoice.models import Invoice


class PaymentNoticeStatusMapping:
    _IMIS_ACTIVE_PAYED = Invoice.Status.PAID.value
    _IMIS_ACTIVE_VALIDATED = Invoice.Status.VALIDATED.value
    _IMIS_ACTIVE_SUSPENDED = Invoice.Status.SUSPENDED.value
    _FHIR_ACTIVE_PAYED = R4PaymentNoticeConfig.get_fhir_payment_notice_status_active()
    _FHIR_ACTIVE_VALIDATED = R4PaymentNoticeConfig.get_fhir_payment_notice_status_active()
    _FHIR_ACTIVE_SUSPENDED = R4PaymentNoticeConfig.get_fhir_payment_notice_status_active()

    _IMIS_CANCELLED = Invoice.Status.CANCELLED.value
    _FHIR_CANCELLED = R4PaymentNoticeConfig.get_fhir_payment_notice_status_cancelled()

    _IMIS_DRAFT = Invoice.Status.DRAFT.value
    _FHIR_DRAFT = R4PaymentNoticeConfig.get_fhir_payment_notice_status_draft()

    _IMIS_ENTERED_IN_ERROR = Invoice.Status.DELETED.value
    _FHIR_ENTERED_IN_ERROR = R4PaymentNoticeConfig.get_fhir_payment_notice_status_entered_in_error()

    to_fhir_status = {
        _IMIS_ACTIVE_PAYED: _FHIR_ACTIVE_PAYED,
        _IMIS_ACTIVE_VALIDATED: _FHIR_ACTIVE_VALIDATED,
        _IMIS_ACTIVE_SUSPENDED: _FHIR_ACTIVE_SUSPENDED,
        _IMIS_CANCELLED: _FHIR_CANCELLED,
        _IMIS_DRAFT: _FHIR_DRAFT,
        _IMIS_ENTERED_IN_ERROR: _FHIR_ENTERED_IN_ERROR
    }

    to_imis_status = {
        _FHIR_ACTIVE_PAYED: _IMIS_ACTIVE_PAYED,
        _FHIR_CANCELLED: _IMIS_CANCELLED,
        _FHIR_DRAFT: _IMIS_DRAFT,
        _FHIR_ENTERED_IN_ERROR: _IMIS_ENTERED_IN_ERROR
    }



from api_fhir_r4.configurations import LocationConfiguration


class R4LocationConfig(LocationConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_location_type = cfg['R4_fhir_location_site_type']
        cls.get_config().R4_fhir_location_physical_type = cfg['R4_fhir_location_physical_type']
        cls.get_config().R4_fhir_hf_service_type = cfg['R4_fhir_hf_service_type']
        cls.get_config().R4_fhir_location_status_codes = cfg['R4_fhir_location_status_codes']
        cls.get_config().R4_fhir_location_subscribe_signal = cfg['R4_fhir_location_subscribe_signal']

    @classmethod
    def get_fhir_location_site_type_system(cls):
        return cls.get_config_attribute("R4_fhir_location_type")\
            .get('system', "http://hl7.org/fhir/v3/ServiceDeliveryLocationRoleType/vs.html")

    @classmethod
    def get_fhir_code_for_hospital(cls):
        return cls.get_config_attribute("R4_fhir_location_type").get('fhir_code_for_hospital', "HOSP")

    @classmethod
    def get_fhir_code_for_dispensary(cls):
        return cls.get_config_attribute("R4_fhir_location_type").get('fhir_code_for_dispensary', "COMM")

    @classmethod
    def get_fhir_code_for_health_center(cls):
        return cls.get_config_attribute("R4_fhir_location_type").get('fhir_code_for_health_center', "OF")

    @classmethod
    def get_fhir_location_area_type_system(cls):
        return cls.get_config_attribute("R4_fhir_location_type")\
            .get('system', "https://openimis.atlassian.net/wiki/spaces/OP/pages/1389297887/FHIR+r4+-+Location")

    @classmethod
    def get_fhir_code_for_region(cls):
        return cls.get_config_attribute("R4_fhir_location_type").get('fhir_code_for_region', "R")

    @classmethod
    def get_fhir_code_for_district(cls):
        return cls.get_config_attribute("R4_fhir_location_type").get('fhir_code_for_district', "D")

    @classmethod
    def get_fhir_code_for_ward(cls):
        return cls.get_config_attribute("R4_fhir_location_type").get('fhir_code_for_ward', "W")

    @classmethod
    def get_fhir_code_for_village(cls):
        return cls.get_config_attribute("R4_fhir_location_type").get('fhir_code_for_village', "V")

    @classmethod
    def get_fhir_location_physical_type_system(cls):
        return cls.get_config_attribute("R4_fhir_location_physical_type")\
            .get('system', "https://openimis.github.io/openimis_fhir_r4_ig/CodeSystem/location-type")

    @classmethod
    def get_fhir_code_for_site(cls):
        return cls.get_config_attribute("R4_fhir_location_type").get('fhir_code_for_site', "si")
    
    @classmethod
    def get_fhir_code_for_area(cls):
        return cls.get_config_attribute("R4_fhir_location_type").get('fhir_code_for_area', "area")

# This should be part organisation because this is not related to the buiilding

    @classmethod
    def get_fhir_hf_service_type_system(cls):
        return cls.get_config_attribute("R4_fhir_hf_service_type")\
            .get('system', "http://hl7.org/fhir/valueset-service-type.html")

    @classmethod
    def get_fhir_code_for_in_patient(cls):
        return cls.get_config_attribute("R4_fhir_hf_service_type").get('fhir_code_for_in_patient', "I")

    @classmethod
    def get_fhir_code_for_out_patient(cls):
        return cls.get_config_attribute("R4_fhir_hf_service_type").get('fhir_code_for_out_patient', "O")

    @classmethod
    def get_fhir_code_for_both(cls):
        return cls.get_config_attribute("R4_fhir_hf_service_type").get('fhir_code_for_both', "B")

    @classmethod
    def get_fhir_code_for_active(cls):
        return cls.get_config_attribute("R4_fhir_location_status_codes").get('fhir_code_for_active', 'active')

    @classmethod
    def get_fhir_code_for_inactive(cls):
        return cls.get_config_attribute("R4_fhir_location_status_codes").get('fhir_code_for_inactive', 'inactive')

    @classmethod
    def get_subscribe_location_signal(cls):
        return cls.get_config_attribute("R4_fhir_location_subscribe_signal")



import datetime

from django.db.models import OuterRef, Exists
from rest_framework import viewsets, status
from rest_framework.response import Response

from api_fhir_r4.converters import OperationOutcomeConverter
from api_fhir_r4.mixins import MultiIdentifierRetrieverMixin, MultiIdentifierUpdateMixin
from api_fhir_r4.model_retrievers import UUIDIdentifierModelRetriever, CHFIdentifierModelRetriever
from api_fhir_r4.permissions import FHIRApiInsureePermissions
from api_fhir_r4.serializers import PatientSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from claim.models import Claim
from insuree.models import Insuree


class InsureeViewSet(BaseFHIRView, MultiIdentifierRetrieverMixin,
                     MultiIdentifierUpdateMixin, viewsets.ModelViewSet):
    retrievers = [UUIDIdentifierModelRetriever, CHFIdentifierModelRetriever]
    serializer_class = PatientSerializer
    permission_classes = (FHIRApiInsureePermissions,)

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        ref_date_str = request.GET.get('refDate')
        claim_date = request.GET.get('claimDateFrom')
        identifier = request.GET.get("identifier")

        if identifier:
            return self.retrieve(request, *args, **{**kwargs, 'identifier': identifier})
        else:
            queryset = queryset.filter(validity_to__isnull=True).order_by('validity_from')
            if ref_date_str is not None:
                try:
                    ref_date = datetime.datetime.strptime(ref_date_str, "%Y-%m-%d").date()
                    queryset = queryset.filter(validity_from__gte=ref_date)
                except ValueError:
                    pass
            if claim_date is not None:
                try:
                    claim_parse_dated = datetime.datetime.strptime(claim_date, "%Y-%m-%d").date()
                except ValueError:
                    result = OperationOutcomeConverter \
                        .build_for_400_bad_request("claimDateFrom should be in dd-mm-yyyy format")
                    return Response(result.dict(), status.HTTP_400_BAD_REQUEST)
                has_claim_in_range = Claim.objects \
                    .filter(date_claimed__gte=claim_parse_dated) \
                    .filter(insuree_id=OuterRef("id")) \
                    .values("id")
                queryset = queryset \
                    .annotate(has_claim_in_range=Exists(has_claim_in_range)) \
                    .filter(has_claim_in_range=True)

        serializer = PatientSerializer(self.paginate_queryset(queryset), many=True)
        return self.get_paginated_response(serializer.data)

    def get_queryset(self):
        queryset = Insuree.get_queryset(None, self.request.user) \
            .select_related('gender') \
            .select_related('photo') \
            .select_related('family__location')

        return ValidityFromRequestParameterFilter(self.request).filter_queryset(queryset)
